package ru.dbotthepony.kstarbound.server

import com.google.gson.JsonObject
import io.netty.channel.ChannelHandlerContext
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.io.ByteKey
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.defs.WarpAlias
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.network.ConnectionSide
import ru.dbotthepony.kstarbound.network.ConnectionType
import ru.dbotthepony.kstarbound.network.IServerPacket
import ru.dbotthepony.kstarbound.network.packets.ClientContextUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.PlayerWarpResultPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.ServerDisconnectPacket
import ru.dbotthepony.kstarbound.server.world.ServerWorldTracker
import ru.dbotthepony.kstarbound.server.world.WorldStorage
import ru.dbotthepony.kstarbound.server.world.LegacyWorldStorage
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import java.util.HashMap
import kotlin.properties.Delegates

// serverside part of connection
class ServerConnection(val server: StarboundServer, type: ConnectionType) : Connection(ConnectionSide.SERVER, type) {
	var world: ServerWorld? = null
	var tracker: ServerWorldTracker? = null
	var worldStartAcknowledged = false

	// packets which interact with world must be
	// executed on world's thread
	fun enqueue(task: ServerWorld.() -> Unit) = tracker?.enqueue(task)

	lateinit var shipWorld: ServerWorld
		private set

	init {
		connectionID = server.channels.nextConnectionID()

		rpc.add("team.fetchTeamStatus") {
			JsonObject()
		}
	}

	override fun toString(): String {
		val channel = if (hasChannel) channel.remoteAddress().toString() else "<no channel>"
		val ship = if (::shipWorld.isInitialized) shipWorld.toString() else "<no shipworld>"
		return "ServerConnection[ID=$connectionID channel=$channel / $ship]"
	}

	private val shipChunks = HashMap<ByteKey, KOptional<ByteArray>>()
	private val modifiedShipChunks = ObjectOpenHashSet<ByteKey>()
	var shipChunkSource by Delegates.notNull<WorldStorage>()
		private set

	override fun setupLegacy() {
		super.setupLegacy()
		shipChunkSource = LegacyWorldStorage.memory(shipChunks)
	}

	override fun setupNative() {
		super.setupNative()
		shipChunkSource = WorldStorage.EMPTY
	}

	fun receiveShipChunks(chunks: Map<ByteKey, KOptional<ByteArray>>) {
		check(shipChunks.isEmpty()) { "Already has ship chunks" }
		shipChunks.putAll(chunks)
	}

	override fun flush() {
		if (isConnected) {
			val entries = rpc.write()

			if (entries != null || modifiedShipChunks.isNotEmpty()) {
				channel.write(ClientContextUpdatePacket(
					entries ?: listOf(),
					KOptional(modifiedShipChunks.associateWith { shipChunks[it]!! }),
					KOptional(ByteArrayList())))

				modifiedShipChunks.clear()
			}
		}

		super.flush()
	}

	override fun onChannelClosed() {
		playerEntity = null

		super.onChannelClosed()
		server.channels.freeConnectionID(connectionID)
		server.channels.connections.remove(this)
		server.freeNickname(nickname)

		announceDisconnect("Connection to remote host is lost.")

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (countedTowardsPlayerCount) {
			countedTowardsPlayerCount = false
			server.channels.decrementPlayerCount()
		}
	}

	private var announcedDisconnect = false

	private fun announceDisconnect(reason: String) {
		if (!announcedDisconnect && nickname.isNotBlank()) {
			if (reason.isBlank()) {
				server.chat.systemMessage("Player '$nickname' disconnected")
			} else {
				server.chat.systemMessage("Player '$nickname' disconnected ($reason)")
			}

			announcedDisconnect = true
		}
	}

	override fun disconnect(reason: String) {
		announceDisconnect(reason)

		if (channel.isOpen) {
			// send pending updates
			flush()
		}

		tracker?.remove()
		tracker = null

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (channel.isOpen) {
			// say goodbye
			channel.write(ServerDisconnectPacket(reason))
			channel.flush()
			channel.close()
		}
	}

	override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
		if (msg is IServerPacket) {
			try {
				msg.play(this)
			} catch (err: Throwable) {
				LOGGER.error("Failed to handle serverbound packet $msg", err)
				disconnect("Incoming packet caused an exception: $err")
			}
		} else {
			LOGGER.error("Unknown serverbound packet type $msg")
			disconnect("Unknown serverbound packet type $msg")
		}
	}

	private var countedTowardsPlayerCount = false

	override fun inGame() {
		server.chat.systemMessage("Player '$nickname' connected")
		countedTowardsPlayerCount = true
		server.channels.incrementPlayerCount()

		if (!isLegacy) {
			server.playerInGame(this)
		} else {
			LOGGER.info("Initializing ship world for $this")

			ServerWorld.load(server, shipChunkSource).thenAccept {
				shipWorld = it
				shipWorld.thread.start()
				send(PlayerWarpResultPacket(true, WarpAlias.OwnShip, false))

				//server.worlds.first().acceptPlayer(this)

				shipWorld.acceptClient(this).exceptionally {
					LOGGER.error("Shipworld of $this rejected to accept its owner", it)
					disconnect("Shipworld rejected player warp request: $it")
				}
			}.exceptionally {
				LOGGER.error("Error while initializing shipworld for $this", it)
				disconnect("Error while initializing shipworld for player: $it")
				null
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
