package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.ints.Int2LongOpenHashMap
import it.unimi.dsi.fastutil.ints.Int2ObjectMaps
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ObjectLinkedOpenHashSet
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.network.packets.ChunkCellsPacket
import ru.dbotthepony.kstarbound.network.IPacket
import ru.dbotthepony.kstarbound.network.packets.EntityCreatePacket
import ru.dbotthepony.kstarbound.network.packets.EntityUpdateSetPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.CentralStructureUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.LegacyTileArrayUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.LegacyTileUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.TileDamageUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.WorldStartPacket
import ru.dbotthepony.kstarbound.server.ServerConnection
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.player.PlayerEntity
import java.io.DataOutputStream
import java.util.HashMap
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.atomic.AtomicBoolean

// couples ServerWorld and ServerConnection together,
// allowing ServerConnection client to track ServerWorld state
class ServerWorldTracker(val world: ServerWorld, val client: ServerConnection) {
	init {
		client.worldStartAcknowledged = false
		client.tracker = this
	}

	var skyVersion = 0L

	private val isRemoved = AtomicBoolean()
	private val tickets = HashMap<ChunkPos, Ticket>()
	private val pendingSend = ObjectLinkedOpenHashSet<ChunkPos>()
	private val tasks = ConcurrentLinkedQueue<ServerWorld.() -> Unit>()
	private val entityVersions = Int2LongOpenHashMap()

	init {
		entityVersions.defaultReturnValue(-1L)
	}

	fun send(packet: IPacket) = client.send(packet)

	// packets which interact with world must be
	// executed on world's thread
	fun enqueue(task: ServerWorld.() -> Unit) {
		tasks.add(task)
	}

	private inner class Ticket(val ticket: ServerWorld.ITicket, val pos: ChunkPos) : IChunkListener {
		override fun onEntityAdded(entity: AbstractEntity) {}
		override fun onEntityRemoved(entity: AbstractEntity) {}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			if (pos !in pendingSend) {
				send(LegacyTileUpdatePacket(pos.tile + Vector2i(x, y), cell.toLegacyNet()))
			}
		}

		override fun onTileHealthUpdate(x: Int, y: Int, isBackground: Boolean, health: TileHealth) {
			// let's hope nothing bad happens from referencing live data
			send(TileDamageUpdatePacket(pos.tileX + x, pos.tileY + y, isBackground, health))
		}
	}

	init {
		val (skyData, skyVersion) = world.sky.networkedGroup.write(isLegacy = client.isLegacy)
		this.skyVersion = skyVersion

		if (client.isLegacy) {
			client.send(WorldStartPacket(
				templateData = Starbound.writeLegacyJson { world.template.toJson() },
				skyData = skyData.toByteArray(),
				weatherData = ByteArray(0),
				playerStart = world.playerSpawnPosition,
				playerRespawn = world.playerSpawnPosition,
				respawnInWorld = world.respawnInWorld,
				dungeonGravity = mapOf(),
				dungeonBreathable = mapOf(),
				protectedDungeonIDs = world.protectedDungeonIDs,
				worldProperties = world.properties.deepCopy(),
				connectionID = client.connectionID,
				localInterpolationMode = false,
			))

			Starbound.writeLegacyJson {
				client.sendAndFlush(CentralStructureUpdatePacket(Starbound.gson.toJsonTree(world.centralStructure)))
			}
		}
	}

	fun isTracking(pos: ChunkPos): Boolean {
		return pos in tickets
	}

	fun tick() {
		if (!client.worldStartAcknowledged)
			return

		if (!client.channel.isOpen) {
			remove() // ???
			return
		}

		run {
			var next = tasks.poll()

			while (next != null) {
				next.invoke(world)
				next = tasks.poll()
			}
		}

		client.playerEntity = world.entities[client.playerID.get()] as? PlayerEntity

		run {
			val newTrackedChunks = ObjectArraySet<ChunkPos>()

			for (region in client.trackingTileRegions()) {
				newTrackedChunks.addAll(world.geometry.tileRegion2Chunks(region))
			}

			val itr = tickets.entries.iterator()

			for ((pos, ticket) in itr) {
				if (pos !in newTrackedChunks) {
					pendingSend.remove(pos)
					ticket.ticket.cancel()
					itr.remove()
				}
			}

			for (pos in newTrackedChunks) {
				if (pos !in tickets) {
					val ticket = world.permanentChunkTicket(pos)
					val thisTicket = Ticket(ticket, pos)
					tickets[pos] = thisTicket
					ticket.listener = thisTicket
					pendingSend.add(pos)
				}
			}
		}

		run {
			val itr = pendingSend.iterator()

			for (pos in itr) {
				val chunk = world.chunkMap[pos] ?: continue

				if (client.isLegacy) {
					send(LegacyTileArrayUpdatePacket(chunk))
					chunk.tileDamagePackets().forEach { send(it) }
				} else {
					send(ChunkCellsPacket(chunk))
				}

				itr.remove()
			}
		}

		for ((id, entity) in world.entities) {
			if (entity.connectionID != client.connectionID && entity is PlayerEntity) {
				if (entityVersions.get(id) == -1L) {
					// never networked
					val initial = FastByteArrayOutputStream()
					entity.writeNetwork(DataOutputStream(initial), client.isLegacy)
					val (data, version) = entity.networkGroup.write(isLegacy = client.isLegacy)

					entityVersions.put(id, version)

					send(EntityCreatePacket(
						entity.type,
						ByteArrayList.wrap(initial.array, initial.length),
						data,
						entity.entityID
					))
				} else {
					val (data, version) = entity.networkGroup.write(remoteVersion = entityVersions.get(id), isLegacy = client.isLegacy)
					entityVersions.put(id, version)
					send(EntityUpdateSetPacket(entity.connectionID, Int2ObjectMaps.singleton(entity.entityID, data)))
				}
			}
		}
	}

	fun remove() {
		if (isRemoved.compareAndSet(false, true)) {
			client.tracker = null
			client.playerEntity = null
			world.players.remove(this)
			tickets.values.forEach { it.ticket.cancel() }

			world.mailbox.execute {
				val itr = world.entities.int2ObjectEntrySet().iterator()

				for ((id, entity) in itr) {
					if (id in client.entityIDRange) {
						entity.remove()
					}
				}
			}
		}
	}
}
