package ru.dbotthepony.kstarbound.world.entities

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.MailboxExecutorService
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.render.LayeredRenderer
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.defs.JsonDriven
import ru.dbotthepony.kstarbound.network.packets.EntityDestroyPacket
import ru.dbotthepony.kstarbound.network.syncher.MasterElement
import ru.dbotthepony.kstarbound.network.syncher.NetworkedGroup
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.LightCalculator
import ru.dbotthepony.kstarbound.world.World
import java.io.DataOutputStream
import java.util.function.Consumer
import kotlin.concurrent.withLock

abstract class AbstractEntity(path: String) : JsonDriven(path) {
	/**
	 * The chunk this entity resides in
	 */
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (innerWorld == null) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (value != null && innerWorld != value.world) {
				throw IllegalArgumentException("$this belongs to $innerWorld, $value belongs to ${value.world}")
			} else if (value == field) {
				return
			}

			val oldChunk = field
			field = value

			world.lock.withLock {
				if (oldChunk == null && value != null) {
					world.orphanedEntities.remove(this)
					value.addEntity(this)
				} else if (oldChunk != null && value == null) {
					world.orphanedEntities.add(this)
					oldChunk.removeEntity(this)
				} else if (oldChunk != null && value != null) {
					value.transferEntity(this, oldChunk)
				}
			}
		}

	abstract val chunkPos: ChunkPos
	abstract val position: Vector2d

	var entityID: Int = 0
		set(value) {
			if (field != 0)
				throw IllegalStateException("Already has Entity ID set (to $field)")

			field = value

			if (value < 0) {
				connectionID = (-value - 1) / 65536 + 1
			}
		}

	var connectionID: Int = 0
		private set

	private val exceptionLogger = Consumer<Throwable> {
		LOGGER.error("Error while executing queued task on $this", it)
	}

	var mailbox = MailboxExecutorService().also { it.exceptionHandler = exceptionLogger }
		private set

	private var innerWorld: World<*, *>? = null

	val world: World<*, *>
		get() = innerWorld ?: throw IllegalStateException("Not in world")

	val isSpawned: Boolean
		get() = innerWorld != null

	abstract val type: EntityType

	/**
	 * If set, then the entity will be discoverable by its unique id and will be
	 * indexed in the stored world.  Unique ids must be different across all
	 * entities in a single world.
	 */
	var uniqueID: String? = null
		protected set

	var description = ""

	/**
	 * Whenever this entity should be removed when chunk containing it is being unloaded
	 *
	 * Returning false will also stop entity from being saved to disk, and render entity orphaned
	 * when chunk containing it will get unloaded
	 */
	open val isApplicableForUnloading: Boolean
		get() = true

	protected open fun onJoinWorld(world: World<*, *>) { }
	protected open fun onRemove(world: World<*, *>) { }

	val networkGroup = MasterElement(NetworkedGroup())
	abstract fun writeNetwork(stream: DataOutputStream, isLegacy: Boolean)

	fun joinWorld(world: World<*, *>) {
		if (innerWorld != null)
			throw IllegalStateException("Already spawned (in world $innerWorld)")

		if (entityID == 0)
			entityID = world.nextEntityID.incrementAndGet()

		world.ensureSameThread()

		check(!world.entities.containsKey(entityID)) { "Duplicate entity ID: $entityID" }

		if (mailbox.isShutdown)
			mailbox = MailboxExecutorService().also { it.exceptionHandler = exceptionLogger }

		innerWorld = world
		world.entities[entityID] = this
		world.orphanedEntities.add(this)
		onJoinWorld(world)
	}

	fun remove() {
		val world = innerWorld ?: throw IllegalStateException("Not in world")
		world.ensureSameThread()

		mailbox.shutdownNow()
		chunk = null
		check(world.entities.remove(entityID) == this) { "Tried to remove $this from $world, but removed something else!" }
		world.orphanedEntities.remove(this)
		onRemove(world)
		world.broadcast(EntityDestroyPacket(entityID, ByteArrayList(), false))
		innerWorld = null
	}

	var isRemote: Boolean = false

	fun tick() {
		tickShared()

		if (isRemote) {
			tickRemote()
		} else {
			tickLocal()
		}
	}

	protected open fun tickShared() {
		mailbox.executeQueuedTasks()
	}

	protected open fun tickRemote() {
		networkGroup.upstream.tickInterpolation(Starbound.TIMESTEP)
	}

	protected open fun tickLocal() {

	}

	open fun render(client: StarboundClient, layers: LayeredRenderer) {

	}

	open fun addLights(lightCalculator: LightCalculator, xOffset: Int, yOffset: Int) {

	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
