package ru.dbotthepony.kstarbound.world.entities

import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.io.BinaryStringCodec
import ru.dbotthepony.kommons.io.IntValueCodec
import ru.dbotthepony.kommons.io.map
import ru.dbotthepony.kommons.io.readKOptional
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.io.writeKOptional
import ru.dbotthepony.kommons.matrix.Matrix3f
import ru.dbotthepony.kommons.util.AABB
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.animation.AnimationDefinition
import ru.dbotthepony.kstarbound.defs.animation.ParticleConfig
import ru.dbotthepony.kstarbound.defs.animation.ParticleFactory
import ru.dbotthepony.kstarbound.fromJson
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.math.PeriodicFunction
import ru.dbotthepony.kstarbound.network.syncher.AABBCodecLegacy
import ru.dbotthepony.kstarbound.network.syncher.AABBCodecNative
import ru.dbotthepony.kstarbound.network.syncher.InternedStringCodec
import ru.dbotthepony.kstarbound.network.syncher.NetworkedGroup
import ru.dbotthepony.kstarbound.network.syncher.NetworkedElement
import ru.dbotthepony.kstarbound.network.syncher.NetworkedMap
import ru.dbotthepony.kstarbound.network.syncher.NetworkedSignal
import ru.dbotthepony.kstarbound.network.syncher.networkedAABBNullable
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedColor
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedEventCounter
import ru.dbotthepony.kstarbound.network.syncher.networkedFixedPoint
import ru.dbotthepony.kstarbound.network.syncher.networkedFloat
import ru.dbotthepony.kstarbound.network.syncher.networkedList
import ru.dbotthepony.kstarbound.network.syncher.networkedPointer
import ru.dbotthepony.kstarbound.network.syncher.networkedSignedInt
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import ru.dbotthepony.kstarbound.network.syncher.networkedUnsignedInt
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.Collections
import kotlin.math.atan2
import kotlin.math.cos
import kotlin.math.sin
import kotlin.math.sqrt


class Animator() {
	class Light {
		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var active by networkedBoolean().also { elements.add(it) }
		var xPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var color by networkedColor().also { elements.add(it) }
		var pointAngle by networkedFixedPoint(0.01).also { elements.add(it); it.interpolator = Interpolator.Linear }

		var anchorPart: String? = null
		var transformationGroups: List<String> = listOf()
		var rotationGroup: String? = null
		var rotationCenter: Vector2d? = null

		var flicker: PeriodicFunction? = null
		var pointLight: Boolean = false
		var pointBeam: Float = 0f
		var beamAmbience: Float = 0f
	}

	enum class SoundSignal {
		PLAY, STOP_ALL;

		companion object {
			val CODEC = IntValueCodec.map({ entries[this] }, { ordinal })
		}
	}

	class Sound {
		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var rangeMultiplier = 0.0

		var soundPool by networkedList(InternedStringCodec).also { elements.add(it) }
		var xPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var volumeTarget by networkedFloat(1.0).also { elements.add(it) }
		var volumeRampTime by networkedFloat(0.0).also { elements.add(it) }
		var pitchMultiplierTarget by networkedFloat(1.0).also { elements.add(it) }
		var pitchMultiplierRampTime by networkedFloat(0.0).also { elements.add(it) }
		var loops by networkedSignedInt().also { elements.add(it) }
		val signals = NetworkedSignal(SoundSignal.CODEC).also { elements.add(it) }
	}

	class Effect(val type: String, val time: Double, val directives: String) {
		val enabled = networkedBoolean()
		var timer: Double = 0.0
	}

	class StateInfo {
		val stateIndex = networkedPointer()
		val startedEvent = networkedEventCounter()
	}

	class RotationGroup {
		var angularVelocity = 0.0
		var rotationCenter = Vector2d.ZERO
		val targetAngle = networkedFloat()
		var currentAngle = 0.0
		val immediateEvent = networkedEventCounter()
	}

	class TransformationGroup {
		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var interpolated = false

		var xTranslation by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yTranslation by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var xScale by networkedFloat(1.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yScale by networkedFloat(1.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var xShear by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yShear by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }

		fun affineTransform(): Matrix3f {
			return Matrix3f.rowMajor(
				(xScale * cos(xShear)).toFloat(), (xScale * sin(xShear)).toFloat(), xTranslation.toFloat(),
				(yScale * sin(yShear)).toFloat(), (yScale * cos(yShear)).toFloat(), yTranslation.toFloat(),
				0f, 0f, 1f
			)
		}

		fun setAffineTransform(value: Matrix3f) {
			xTranslation = value[2, 0].toDouble()
			yTranslation = value[2, 1].toDouble()

			xScale = sqrt(value[0, 0].toDouble() * value[0, 0].toDouble() + value[1, 0].toDouble() * value[1, 0].toDouble())
			yScale = sqrt(value[0, 1].toDouble() * value[0, 1].toDouble() + value[1, 1].toDouble() * value[1, 1].toDouble())

			xShear = atan2(value[1, 0].toDouble(), value[0, 0].toDouble())
			yShear = atan2(value[0, 1].toDouble(), value[1, 1].toDouble())
		}
	}

	class ParticleEmitter {
		data class Config(val count: Int, val offset: Vector2d, val flip: Boolean, val factory: ParticleFactory)

		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var emissionRate by networkedFloat().also { elements.add(it) }
		var burstCount by networkedUnsignedInt().also { elements.add(it) }
		var randomSelectCount by networkedUnsignedInt().also { elements.add(it) }
		var emissionRateVariance = 0.0
		var offsetRegion by networkedAABBNullable().also { elements.add(it) }
		var anchorPart: String? = null
		var transformationGroups: List<String> = listOf()
		var rotationGroup: String? = null
		var rotationCenter: Vector2d? = null

		val particleList = ArrayList<Config>()

		var active by networkedBoolean().also { elements.add(it) }
		val burstEvent = networkedEventCounter().also { elements.add(it); it.ignoreOccurrencesOnLoad = true }

		var timer = 0.0
	}

	val networkGroup = NetworkedGroup()

	private val elements = ArrayList<NetworkedElement>()

	var animatedParts = AnimatedParts()
		private set

	var processingDirectives by networkedString().also { elements.add(it) }
	var zoom by networkedFloat().also { elements.add(it) }
	var isFlipped by networkedBoolean().also { elements.add(it) }
	var flippedRelativeCenterLine by networkedFloat().also { elements.add(it) }
	var animationRate by networkedFloat(1.0).also { elements.add(it); it.interpolator = Interpolator.Linear }

	private val globalTags = NetworkedMap(InternedStringCodec, InternedStringCodec)
	private val partTags = HashMap<String, NetworkedMap<String, String>>()

	private val stateInfo = Object2ObjectAVLTreeMap<String, StateInfo>()
	private val rotationGroups = Object2ObjectAVLTreeMap<String, RotationGroup>()
	private val transformationGroups = Object2ObjectAVLTreeMap<String, TransformationGroup>()
	private val particleEmitters = Object2ObjectAVLTreeMap<String, ParticleEmitter>()
	private val lights = Object2ObjectAVLTreeMap<String, Light>()
	private val sounds = Object2ObjectAVLTreeMap<String, Sound>()
	private val effects = Object2ObjectAVLTreeMap<String, Effect>()

	init {
		setupNetworkElements()
	}

	constructor(config: AnimationDefinition) : this() {
		if (config.animatedParts != null)
			animatedParts = AnimatedParts(config.animatedParts)

		for ((k, v) in config.globalTagDefaults) {
			globalTags[k] = v
		}

		for ((part, tags) in config.partTagDefaults) {
			for ((k, v) in tags) {
				setPartTag(part, k, v)
			}
		}

		for ((k, v) in config.transformationGroups) {
			val group = TransformationGroup()
			transformationGroups[k] = group
			group.interpolated = v.interpolated
		}

		for ((k, v) in config.rotationGroups) {
			val group = RotationGroup()
			group.angularVelocity = v.angularVelocity
			group.rotationCenter = v.rotationCenter
			rotationGroups[k] = group
		}

		for ((k, v) in config.particleEmitters) {
			val emitter = ParticleEmitter()

			emitter.emissionRate = v.emissionRate
			emitter.emissionRateVariance = v.emissionRateVariance
			emitter.offsetRegion = KOptional.ofNullable(v.offsetRegion)
			emitter.anchorPart = v.anchorPart
			emitter.transformationGroups = v.transformationGroups
			emitter.rotationGroup = v.rotationGroup
			emitter.rotationCenter = v.rotationCenter
			emitter.burstCount = v.burstCount
			emitter.randomSelectCount = v.randomSelectCount
			emitter.active = v.active

			for (particle in v.particles) {
				val factory = particle.particle.map({ it.value }, { it }) ?: continue // not a valid particle, too bad.
				emitter.particleList.add(ParticleEmitter.Config(particle.count, particle.offset, particle.flip, factory))
			}

			particleEmitters[k] = emitter
		}

		for ((k, v) in config.lights) {
			val light = Light()

			light.active = v.active
			light.xPosition = v.position.x
			light.yPosition = v.position.y
			light.color = v.color
			light.anchorPart = v.anchorPart
			light.transformationGroups = v.transformationGroups
			light.rotationGroup = v.rotationGroup
			light.rotationCenter = v.rotationCenter
			light.pointAngle = Math.toRadians(v.pointAngle)
			light.pointLight = v.pointLight
			light.pointBeam = v.pointBeam
			light.beamAmbience = v.beamAmbience

			if (v.flickerPeriod != null)
				light.flicker = PeriodicFunction(v.flickerPeriod, v.flickerMinIntensity, v.flickerMaxIntensity, v.flickerPeriodVariance, v.flickerIntensityVariance)

			lights[k] = light
		}

		for ((k, v) in config.sounds) {
			val sound = Sound()

			if (v.isLeft) {
				sound.soundPool = v.left()
			} else {
				val conf = v.right()
				sound.soundPool = conf.pool.map { it.fullPath }
				sound.xPosition = conf.position.x
				sound.yPosition = conf.position.y
				sound.volumeTarget = conf.volume
				sound.volumeRampTime = conf.volumeRampTime
				sound.pitchMultiplierTarget = conf.pitchMultiplier
				sound.pitchMultiplierRampTime = conf.pitchMultiplierRampTime
				sound.rangeMultiplier = conf.rangeMultiplier
			}

			sounds[k] = sound
		}

		for ((k, v) in config.effects) {
			effects[k] = Effect(v.type, v.time, v.directives)
		}

		for (k in animatedParts.stateTypes()) {
			stateInfo[k] = StateInfo()
		}

		for (k in animatedParts.parts()) {
			partTags.computeIfAbsent(k) { NetworkedMap(InternedStringCodec, InternedStringCodec) }
		}

		setupNetworkElements()
	}

	// Every part image can have one or more <tag> directives in it, which if set
	// here will be replaced by the tag value when constructing Drawables.  All
	// Drawables can also have a <frame> tag which will be set to whatever the
	// current state frame is (1 indexed, so the first frame is 1).
	fun setGlobalTag(key: String, value: String) {
		globalTags[key] = value
	}

	fun setPartTag(partName: String, tagKey: String, tagValue: String) {
		var tags = partTags[partName]

		if (tags == null) {
			tags = NetworkedMap(InternedStringCodec, InternedStringCodec)
			partTags[partName] = tags
		}

		tags[tagKey] = tagValue
	}

	private fun setupNetworkElements() {
		networkGroup.clear()
		elements.forEach { networkGroup.add(it) }

		networkGroup.add(globalTags)

		// animated part set
		for (v in animatedParts.parts()) {
			networkGroup.add(partTags[v] ?: throw RuntimeException("Missing animated part $v!"))
		}

		for (v in stateInfo.values) {
			networkGroup.add(v.stateIndex)
			networkGroup.add(v.startedEvent)
		}

		for (v in transformationGroups.values) {
			v.addTo(networkGroup)
		}

		for (v in rotationGroups.values) {
			networkGroup.add(v.targetAngle)
			networkGroup.add(v.immediateEvent)
		}

		for (v in particleEmitters.values) {
			v.addTo(networkGroup)
		}

		for (v in lights.values) {
			v.addTo(networkGroup)
		}

		for (v in sounds.values) {
			v.addTo(networkGroup)
		}

		for (v in effects.values) {
			networkGroup.add(v.enabled)
		}
	}

	companion object {
		// lame
		fun load(path: String): Animator {
			val json = Starbound.loadJsonAsset(path)

			if (json == null) {
				if (missing.add(path)) {
					LOGGER.error("Unable to instance Animator from $path! This very likely gonna make networking code go haywire.")
				}

				return Animator()
			}

			return Animator(Starbound.gson.fromJson(json, AnimationDefinition::class.java))
		}

		private val LOGGER = LogManager.getLogger()
		private val missing = Collections.synchronizedSet(ObjectOpenHashSet<String>())
	}
}
